/*  convert a structure factor file from mmCIF to new mmCIF */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/types.h>
#include "rcsb_math.h"
#include "sf_convert.h"
#include "util.h"

#include "TableFile.h"
#include "CifFile.h"
#include "CifParserBase.h"


extern int TEST, SINGLE;
extern float CELL[6];
extern char SYMM[20], FREERV[20], PDBID[80], AUDIT[512];

SF sf;
void write_sf_info(FILE *fout, char *info, int id);
void compare_refln(char *item);


void clean_up_sf(char *ifile, char *ofile)
//correct the cif tokens, if the ciflist is inside of the SF file.
{
    char  token1[256], token2[256], token[256], token_change[5200];
    char  **line, **refln_list, tmp[215], str[2560];
    int  i=0, j, n=0, n1=0, nline, n2=0, n3=0, nlist=0;
    FILE *fp;


//the list below are the wrong cif token in the mmcif file
    static const char *ciflist[] = 
        {
" _refln.A_cal                              _refln.A_calc ",
" _refln.B_cal                              _refln.B_calc ",
" _refln.resolution                         _refln.d_spacing ",
" _refln.Fc_au                              _refln.F_calc ",
" _refln.F_sigma                            _refln.F_meas_sigma ",
" _refln.F_meas_au_sigma                    _refln.F_meas_sigma_au ",
" _refln.F_meas_siygma_au                   _refln.F_meas_sigma_au ",
" _refln.F_squared_meas_sigma               _refln.F_squared_sigma ",
" _refln.I_meas_au                          _refln.intensity_meas ",
" _refln.intensity                          _refln.intensity_meas ",
" _refln.intensity_meas_au                  _refln.intensity_meas ",
" _refln.intensity_sigm                     _refln.intensity_sigma ",
" _refln.I_meas_sigma_au                    _refln.intensity_sigma ",
" _refln.intensity_sigma_au                 _refln.intensity_sigma ",
" _refln.intensity_meas_sigma               _refln.intensity_sigma ",
" _refln.intensity_meas_sigma_au            _refln.intensity_sigma ",
" _refln.dfano                              _refln.pdbx_anom_difference ",
" _refln.DF_anomalous                       _refln.pdbx_anom_difference ",
" _refln.ccp4_phase_anom                    _refln.pdbx_anom_difference ",
" _refln.pdbx_phase_anom                    _refln.pdbx_anom_difference ",
" _refln.ccp4_anomalous_diff                _refln.pdbx_anom_difference ",
" _refln.ccp4_SAD_phase_anom                _refln.pdbx_anom_difference ",
" _refln.pdbx_anomalous_diff                _refln.pdbx_anom_difference ",
" _refln.pdbx_SAD_phase_anom                _refln.pdbx_anom_difference ",
" _refln.pdbx_anomalous_meas_au             _refln.pdbx_anom_difference ",
" _refln.I_ano_meas                         _refln.pdbx_anom_difference ",
" _refln.I_ano_sigma                        _refln.pdbx_anom_difference_sigma ",
" _refln.DF_anomalous_sigma                 _refln.pdbx_anom_difference_sigma ",
" _refln.ccp4_phase_anom_sigma              _refln.pdbx_anom_difference_sigma ",
" _refln.pdbx_phase_anom_sigma              _refln.pdbx_anom_difference_sigma ",
" _refln.ccp4_anomalous_diff_sigma          _refln.pdbx_anom_difference_sigma ",
" _refln.ccp4_SAD_phase_anom_sigma          _refln.pdbx_anom_difference_sigma ",
" _refln.pdbx_SAD_phase_anom_sigma          _refln.pdbx_anom_difference_sigma ",
" _refln.pdbx_anomalous_meas_sigma_au       _refln.pdbx_anom_difference_sigma ",
" _refln.F-_meas_au                         _refln.pdbx_F_minus ",
" _refln.pdbx_F_meas_minus                  _refln.pdbx_F_minus ",
" _refln.ccp4_F_meas_minus_au               _refln.pdbx_F_minus ",
" _refln.pdbx_F_meas_minus_au               _refln.pdbx_F_minus ",
" _refln.ccp4_SAD_F_meas_minus_au           _refln.pdbx_F_minus ",
" _refln.F-_meas_sigma_au                   _refln.pdbx_F_minus_sigma ",
" _refln.pdbx_F_meas_minus_sigma            _refln.pdbx_F_minus_sigma  ",
" _refln.ccp4_F_meas_minus_sigma_au         _refln.pdbx_F_minus_sigma ",
" _refln.pdbx_F_meas_minus_sigma_au         _refln.pdbx_F_minus_sigma ",
" _refln.ccp4_SAD_F_meas_minus_sigma_au     _refln.pdbx_F_minus_sigma ",
" _refln.F+_meas_au                         _refln.pdbx_F_plus ",
" _refln.pdbx_F_meas_plus                   _refln.pdbx_F_plus ",
" _refln.ccp4_F_meas_plus_au                _refln.pdbx_F_plus ",
" _refln.pdbx_F_meas_plus_au                _refln.pdbx_F_plus ",
" _refln.ccp4_SAD_F_meas_plus_au            _refln.pdbx_F_plus ",
" _refln.F+_meas_sigma_au                   _refln.pdbx_F_plus_sigma ",
" _refln.pdbx_F_meas_plus_sigma             _refln.pdbx_F_plus_sigma ",
" _refln.ccp4_F_meas_plus_sigma_au          _refln.pdbx_F_plus_sigma ",
" _refln.pdbx_F_meas_plus_sigma_au          _refln.pdbx_F_plus_sigma ",
" _refln.ccp4_SAD_F_meas_plus_sigma_au      _refln.pdbx_F_plus_sigma ",
" _refln.pdbx_HLA                           _refln.pdbx_HL_A_iso ",
" _refln.ccp4_HL_A_iso                      _refln.pdbx_HL_A_iso ",
" _refln.ccp4_SAD_HL_A_iso                  _refln.pdbx_HL_A_iso ",
" _refln.pdbx_SAD_HL_A_iso                  _refln.pdbx_HL_A_iso ",
" _refln.pdbx_HLB                           _refln.pdbx_HL_B_iso ",
" _refln.ccp4_HL_B_iso                      _refln.pdbx_HL_B_iso ",
" _refln.ccp4_SAD_HL_B_iso                  _refln.pdbx_HL_B_iso ",
" _refln.pdbx_SAD_HL_B_iso                  _refln.pdbx_HL_B_iso ",
" _refln.pdbx_HLC                           _refln.pdbx_HL_C_iso ",
" _refln.ccp4_HL_C_iso                      _refln.pdbx_HL_C_iso ",
" _refln.ccp4_SAD_HL_C_iso                  _refln.pdbx_HL_C_iso ",
" _refln.pdbx_SAD_HL_C_iso                  _refln.pdbx_HL_C_iso ",
" _refln.pdbx_HLD                           _refln.pdbx_HL_D_iso ",
" _refln.ccp4_HL_D_iso                      _refln.pdbx_HL_D_iso ",
" _refln.ccp4_SAD_HL_D_iso                  _refln.pdbx_HL_D_iso ",
" _refln.pdbx_SAD_HL_D_iso                  _refln.pdbx_HL_D_iso ",
" _refln.ccp4_I_minus                       _refln.pdbx_I_minus ",
" _refln.ccp4_I_minus_sigma                 _refln.pdbx_I_minus_sigma ",
" _refln.ccp4_I_plus                        _refln.pdbx_I_plus ",
" _refln.ccp4_I_plus_sigma                  _refln.pdbx_I_plus_sigma ",
" _refln.pahse_calc                         _refln.phase_calc ",
" _refln.F.phase_calc                       _refln.phase_calc ",
" _refln.phase_au                           _refln.phase_meas ",
" _refln.pdbx_fom_weighted_fmap             _refln.pdbx_FWT ",
" _refln.statu                              _refln.status ",
" _refln.F_status                           _refln.status ",
" _refln.status_au                          _refln.status ",
" _refln.R_free_flag                        _refln.status ",
" _refln.observed_status                    _refln.status ",
" _refln.wavelength_di                      _refln.wavelength_id ",
" _refln.index_l>                           _refln.index_l ",
" _refln_index_h                            _refln.index_h ",
" _refln_index_k                            _refln.index_k ",
" _refln_index_l                            _refln.index_l ",
" _refln_F_squared_meas                     _refln.F_squared_meas ",
" _refln_F_squared_sigma                    _refln.F_squared_sigma ",
" _refln_F_calc                             _refln.F_calc ",
" _refln_phase_calc                         _refln.phase_calc ",
" _diffrn_radiation.pdbx_wavelength        _diffrn_radiation_wavelength.wavelength ",
" _diffrn_radiation.diffrn_id              _diffrn_radiation_wavelength.id ",
" _refln.fiber_coordinate                  _refln.pdbx_fiber_coordinate ",
" _refln.fiber_F_meas_au                   _refln.pdbx_fiber_F_meas_au ",
" _refln.fiber_layer                       _refln.pdbx_fiber_layer ",
"END",

"_refln.F_meas  _refln.intensity_meas ",
"_refln.F_meas_sigma  _refln.intensity_sigma",
"_refln.F_meas_au  _refln.intensity_meas ",
"_refln.F_meas_sigma_au  _refln.intensity_sigma",
"_refln.intensity_meas  _refln.F_meas_au",
"_refln.intensity_sigma  _refln.F_meas_sigma_au",

"END"
        };

   
        
/*    
    sprintf(excute, "/bin/csh  -c \" /usr/bin/perl  -i -p -e \
's/ccp4_SAD_/pdbx_/g; s/CCP4_/pdbx_/g; s/ccp4_/pdbx_/g;' %s\" ", ofile);
    system(excute);
*/
    
    refln_list=set_memory_2d(200, 90);
    
    line=get_lines_from_file(ifile, &nline);
    if( (fp = fopen(ofile, "w")) ==NULL || nline<10 ){        
        printf("Error! Can not open the file (%s) or no written permission! \n",ifile);
        return ;
    }
    
    strcpy(token_change,"");
    for(i=0; i<nline; i++){
        strcpy(str, line[i]);
        
        if(strstr_case(str, "data_")){
            if(strchr(str, '/')) replace_char(str, '/', '_');
            n2=0;

            
        }else if(strstr_case(str,"_symmetry_equiv.") ){
            if(i>=nline-1) continue;
            sscanf(line[i+1], "%s", tmp);
            if(strstr(tmp, "_atom_sites.")){
                printf("Error! Table 'symmetry_equiv' is empty (%s)\n", ifile);
                
                fprintf(fp, "%s 1 \n", str);
                continue;
                
            }
            
        }else if(strstr_case(str,"_diffrn_radiation.")){// do not use this
            if(strstr(str,"_diffrn_radiation.pdbx_wavelength")){
                replace_first_str(str, "_diffrn_radiation.pdbx_wavelength",
                                  "_diffrn_radiation_wavelength.wavelength", tmp);
            }else{
                replace_first_str(str, "_diffrn_radiation.", "_diffrn_radiation_wavelength.", tmp);
            }
//            printf("Note: changed _diffrn_radiation to _diffrn_radiation_wavelength\n");
            
            
            fprintf(fp, "%s", tmp);
            continue;
            
        }else if(strstr_case(str,"_refln.")){//current cif parser failed to parse loop
            n1=sscanf(str,"%s %d", tmp, &n);
            if(n1 == 2){
                if (!strncmp_case(tmp,"_refln.", 7) && n>0) {
                    printf("Error in cif! deleting line %s\n", str);
                    continue;
                }
                
            }else if(n1 == 1  && !strncmp(tmp,"_refln.", 7) ){
                compare_refln(tmp);
                
                
                n2++;
                if(n2 == 1) get_refln_list(i, nline, line, refln_list, &nlist);
                               
                sscanf(str, "%s", token1);
                rid_of_front_end_space(token1);
               
                for(j=0; j<500; j++){
                    if(!strcmp_case(ciflist[j], "END"))break;
                    sscanf(ciflist[j], "%s %s", token2, token);
                    if(!strcmp_case(token1, token2)){ //wrong token
                        n3=check_refln_list(token, nlist, refln_list);
                        if(n3==0){ //token not in the refln_list
                            sprintf(tmp,"%s replaced by %s;", token2, token);
                            printf("Warning (%s): %s\n", ifile, tmp);
                            strcat(token_change,tmp);
                            replace_first_str(str,token2, token, str);
                        }else{
                            printf("Error! The item %s can not be replaced by %s\n", token2, token);
                        }
                        
                        break;
                    }
                }
            }
            
        }
        fprintf(fp, "%s", str);
    }
    if(nline>0) free_memory_2d(line, nline);
    fclose(fp);
    
    
}

void get_refln_list(int start, int nline, char **line, char **refln_list, int *nlist)
//get the list of _refln. to see if the replaced item is in the list

{
    char str[256];
    int i, n=0, m=0;

    for(i=start; i<nline; i++){
        if(n++> 50) break;
        strcpy(str, line[i]);
        rid_of_front_end_space(str);
        if(!strncmp(str,"_refln.", 7)){
            strcpy(refln_list[m], str);
            m++;
        }
    }
    *nlist= m;
}

int check_refln_list(char *token, int nlist, char **refln_list)
//Use the list of _refln. to see if the replaced item is in the list

{
    int i, n=0;
        
    for(i=0; i<nlist; i++){
        if(!strcmp_case(token, refln_list[i])){
            n=1;
            break;
        }
    }
    return n;

}

void compare_refln(char *item)
{
    int j, n;
    
    static const char *refln[]={
"_refln.A_calc",
"_refln.A_calc_au",
"_refln.A_meas",
"_refln.A_meas_au",
"_refln.B_calc",
"_refln.B_calc_au",
"_refln.B_meas",
"_refln.B_meas_au",
"_refln.class_code",
"_refln.crystal_id",
"_refln.d_spacing",
"_refln.F_calc",
"_refln.F_calc_au",
"_refln.F_meas",
"_refln.F_meas_au",
"_refln.F_meas_sigma",
"_refln.F_meas_sigma_au",
"_refln.F_squared_calc",
"_refln.F_squared_meas",
"_refln.F_squared_sigma",
"_refln.fom",
"_refln.include_status",
"_refln.index_h",
"_refln.index_k",
"_refln.index_l",
"_refln.intensity_calc",
"_refln.intensity_meas",
"_refln.intensity_sigma",
"_refln.mean_path_length_tbar",
"_refln.pdbx_anom_difference",
"_refln.pdbx_anom_difference_sigma",
"_refln.pdbx_DELFWT",
"_refln.pdbx_DELPHWT",
"_refln.pdbx_diffrn_id",
"_refln.pdbx_F_calc_part_solvent",
"_refln.pdbx_F_calc_with_solvent",
"_refln.pdbx_F_minus",
"_refln.pdbx_F_minus_sigma",
"_refln.pdbx_F_plus",
"_refln.pdbx_F_plus_sigma",
"_refln.pdbx_fiber_coordinate",
"_refln.pdbx_fiber_F_meas_au",
"_refln.pdbx_fiber_layer",
"_refln.pdbx_FWT",
"_refln.pdbx_HL_A_iso",
"_refln.pdbx_HL_B_iso",
"_refln.pdbx_HL_C_iso",
"_refln.pdbx_HL_D_iso",
"_refln.pdbx_I_minus",
"_refln.pdbx_I_minus_sigma",
"_refln.pdbx_I_plus",
"_refln.pdbx_I_plus_sigma",
"_refln.pdbx_phase_calc_part_solvent",
"_refln.pdbx_phase_calc_with_solvent",
"_refln.pdbx_PHWT",
"_refln.phase_calc",
"_refln.phase_meas",
"_refln.refinement_status",
"_refln.scale_group_code",
"_refln.sint_over_lambda",
"_refln.status",
"_refln.symmetry_epsilon",
"_refln.symmetry_multiplicity",
"_refln.wavelength",
"_refln.wavelength_id",
"END"
    };
    n=0;
    for(j=0; j<500; j++){
        if(!strcmp_case(refln[j], "END"))break;
        if(!strcmp_case(item, refln[j])){
            n=1;
        }
    }
    if (n==0){
        printf("Error: %s is not in the pdbx dictionary.\n", item);
    }
}

void cif2cif_sf_all(char *iFile, char *sffile_new, char *pdb_id, int key)
{
    string diags, blockId;
    char ofile[512], fname[2000], **line, block_id[80];
    int  i=0, numBlocks, nstr,  c;
    FILE *fp, *fout, *finfo;
    
    CifFile * fobjR = NULL;

    printf("Doing mmcif to mmcif conversion..\n");

    sprintf(ofile, "%s_new", iFile);
    clean_up_sf(iFile, ofile); 
    

    strcpy(fname, "");
    finfo = fopen("sf_information.txt", "w"); //put inform there
    
    if((fout = fopen(sffile_new, "w"))==NULL){
        open_file_error(sffile_new, "cif2cif_sf_all");
    }

    
    CifParser* cifParserR = NULL;
    fobjR = new CifFile(true);
    cifParserR = new CifParser(fobjR, fobjR->GetVerbose());
    cifParserR->Parse(ofile, diags);
    delete(cifParserR);

    if (!diags.empty()) fprintf(stderr, "Diags for file %s: %s\n",ofile,diags.c_str());

    vector<string> blockNames;
    fobjR->GetBlockNames(blockNames);
    blockId = fobjR->GetFirstBlockName();
    numBlocks = fobjR->GetNumBlocks();
        
    printf("Number of data blocks = %d\n", numBlocks);
    for(i=0; i<numBlocks;i++){ 
        blockId = blockNames[i];
        strcpy(block_id, blockId.c_str());
        if (!strcmp_case(pdb_id, "xxxx")){
            if (strlen(block_id)>6){
                strncpy(pdb_id, block_id+1, 4);
                pdb_id[4]='\0';
                strcpy(PDBID, pdb_id);
            }
        }
        
        printf("Data_block_id=%s,  pdbid=%s\n", block_id, pdb_id);
//        check_data_block_name(block_id, i);
        
        write_hkl_values(fout,finfo, fname, fobjR, i, numBlocks, pdb_id, key);
        if(key==0 && i==0) { //only first block and give F if it is not there.
            fclose(fout);
            delete_file(ofile);
            return;
        }
    }
    if (fobjR) delete fobjR;
    
    line = string_token(fname, "," , &nstr);
//    for(i=0; i<nstr; i++){ printf("FILE=%d || %s\n", i+1,line[i]); }
//    column=0;
//    array_sort_by_column(line, column, 2, nstr);

    rewind(fout);
    for(i=0; i<nstr; i++){
        sscanf(line[i], "%*s%s",fname);
        
        if(i==0){
                //printf("Use SORTED FILE=%d || %s for validation\n",i+1, line[i]);
            fobjR = new CifFile(true);
            cifParserR = new CifParser(fobjR, fobjR->GetVerbose());
            cifParserR->Parse(fname, diags);
            delete(cifParserR);
            
            write_sf_4_validate(iFile, fobjR);
            
            if (fobjR) delete fobjR;
            
        }
           
        fp=fopen(fname, "r");
        while((c=getc(fp))!=EOF){ // write to given file name
            fprintf(fout,"%c",c);
        }
        fclose(fp);
        
        delete_file(fname);
    }
    fclose(fout);
    fclose(finfo);
    printf("SF information is written in file=sf_information.txt\n");
    delete_file(ofile);
}

void check_data_block_name(char *block_id, int num)
{
    int n=0, prob=0;

    n=strlen(block_id);

    
    if (num==0){
        if (n != 7 || block_id[0] !='r' || !strstr(block_id +5, "sf")){
            prob=1;
        }
        
    }else {
        if (n != 8 || block_id[0] !='r' || !strstr(block_id +6, "sf")){
            prob=1;
        }else if(num<25 && (block_id[5]<65 || block_id[5]>90)){
            prob=1;
        }
    }
    if (prob==1){
        printf("Warning: ID (%s) in block=%d is not standard (r????sf, r????Asf ...).\n", block_id, num);
    }
    
}

    
void write_amplitude(FILE *fout, int i, char *I , char **sI, float sf_defaut)
{
    float f=0, sf=0;
    

    if(atof(I)>0){
        f=sqrt(atof(I));
        if(sI && atof(sI[i])>0){
            sf=atof(sI[i])/(2.0*f);
        }else
            sf=sf_defaut;
    }
    fprintf(fout, " %8.2f %6.2f ", f, sf);

}

void other_to_f(int i, float &F, float &Fs)
// key=0,1
{
    float sp1=0, sp2=0, I=0, Is=0;
    
    F = 0;
    Fs = 0.01;
    
    if(sf.Io){ // Io
        if(!strchr(sf.Io[i], '?')){
            I=atof(sf.Io[i]);
            
            if( I>0 ){
                F=sqrt(I);
                if(sf.sIo) Fs= atof(sf.sIo[i])/(2.0*F);
            }
        }
        
    }else if(sf.sI_plus && sf.sI_minus){ // I_plus
        sp1=1;
        sp2=1;
        (strchr(sf.sI_plus[i],'?'))? sp1=-1 : sp1=atof(sf.sI_plus[i]);
        (strchr(sf.sI_minus[i],'?')) ? sp2=-1 : sp2=atof(sf.sI_minus[i]);
            
        I = 0;
        Is=0;
        if(sp1<0 && sp2<0){
            I = 0;
            Is=0;
        }else if(sp1>0 && sp2<0){
            I=atof(sf.I_plus[i]);
            Is=atof(sf.sI_plus[i]);
            
        }else if(sp1<0 && sp2>0){
            I=atof(sf.I_minus[i]);
            Is=atof(sf.sI_minus[i]);
        
        }else if(sp1>0 && sp2>0){
            I = (atof(sf.I_plus[i])+atof(sf.I_minus[i]))/2.0;
            Is= (atof(sf.sI_plus[i])+atof(sf.sI_minus[i]))/2.0;
        }
        if(I>0) {
            F=sqrt(I);
            Fs=Is/(2.0*F);
        }
        
    }else if(sf.F_plus && sf.F_minus){ // F_plus
        
        if(strchr(sf.F_plus[i], '?') && strchr(sf.F_minus[i], '?')){
            F = 0;
            Fs=0;
        }else if(!strchr(sf.F_plus[i], '?') && strchr(sf.F_minus[i], '?')){
            F=atof(sf.F_plus[i]);
            Fs=atof(sf.sF_plus[i]);
        }else if(strchr(sf.F_plus[i], '?') && !strchr(sf.F_minus[i], '?')){
            F=atof(sf.F_minus[i]);
            Fs=atof(sf.sF_minus[i]);
        
        }else if(!strchr(sf.F_plus[i], '?') && !strchr(sf.F_minus[i], '?')){
            F=(atof(sf.F_plus[i])+atof(sf.F_minus[i]))/2.0;
            Fs=(atof(sf.sF_plus[i])+atof(sf.sF_minus[i]))/2.0;
        }
    }
    
    
}

             
void write_sf_4_validate(char *fname,  CifFile *fobjR)
// write a sf file for validation, only for F, sF.
{
    char *flag=NULL, block_name[80];
    long i=0, n=0;
    float F=0, Fs=0,sig_defaut=0.01;
    FILE *fout=NULL;

    
    parse_sf_values(fobjR, 0);
    n=sf.nref;
    
    flag=new char[n+10];
    
    for(i=0; i<n; i++){
        if(sf.status)
            flag[i]=sf.status[i][0];
        else
            flag[i]='o';       
    }
    
    if(n<=0){
        printf("\nError! Number of reflections for validation set =%ld\n",  n);
        return;
    }

    printf("\nNumber of reflections for validation set =%ld  \n",  n);
    fout=fopen("SF_4_validate.cif", "w"); //auto-refine use this name

    strcpy(block_name, "cif2cif");
    fprintf(fout, "data_cif2cif\n");
        //sf_related_head(fout, block_name, block_name);
        
    fprintf(fout, "loop_\n");
    fprintf(fout, " _refln.crystal_id \n");
    fprintf(fout, " _refln.wavelength_id \n");
    fprintf(fout, " _refln.scale_group_code\n");
    fprintf(fout, " _refln.index_h\n");
    fprintf(fout, " _refln.index_k\n");
    fprintf(fout, " _refln.index_l\n");
    fprintf(fout, " _refln.status\n" );
    fprintf(fout, " _refln.F_meas_au \n" );
    fprintf(fout, " _refln.F_meas_sigma_au\n" );

    for(i=0; i<n; i++){
        if(sf.Fo ){
            if(strchr(sf.Fo[i], '?') || strstr(sf.Fo[i], " . ") || atof(sf.Fo[i]) <=0) continue;
            fprintf(fout, "1 1 1 %4s %4s %4s %c ", sf.H[i], sf.K[i], sf.L[i], flag[i]);
                
            fprintf(fout, " %9s ", sf.Fo[i]);
            if(sf.sFo){
                (atof(sf.sFo[i])>sig_defaut)? fprintf(fout," %s\n",sf.sFo[i]):fprintf(fout," %.2f\n", sig_defaut);
            }else{
                fprintf(fout," %.2f\n", sig_defaut);
            }
            
        }else if(sf.Fo_au){
            if(strchr(sf.Fo_au[i], '?') || strstr(sf.Fo_au[i], " . ")|| atof(sf.Fo_au[i])<=0 ) continue;
            fprintf(fout, "1 1 1 %4s %4s %4s %c ", sf.H[i], sf.K[i], sf.L[i], flag[i]);
                
            fprintf(fout, " %9s ", sf.Fo_au[i]);
            if(sf.sFo_au){
                (atof(sf.sFo_au[i])>sig_defaut)? fprintf(fout," %s\n",sf.sFo_au[i]):fprintf(fout," %.2f\n", sig_defaut);
            }else{
                fprintf(fout," %.2f\n", sig_defaut);
            }
        
        }else if(sf.F_plus ){
            fprintf(fout, "1 1 1 %4s %4s %4s %c ", sf.H[i], sf.K[i], sf.L[i], flag[i]);
                
            other_to_f(i, F, Fs) ;
            fprintf(fout, " %7.1f %5.1f\n" , F, Fs);
            
        }else if(sf.Io ){
            fprintf(fout, "1 1 1 %4s %4s %4s %c ", sf.H[i], sf.K[i], sf.L[i], flag[i]);
            
            write_amplitude(fout, i, sf.Io[i], sf.sIo, sig_defaut);
            fprintf(fout," \n");

        }else if(sf.F2o){
            fprintf(fout, "1 1 1 %4s %4s %4s %c ", sf.H[i], sf.K[i], sf.L[i], flag[i]);
            
            write_amplitude(fout, i, sf.F2o[i], sf.sF2o, sig_defaut);
            fprintf(fout," \n");
            
        }else if(sf.I_plus){
            fprintf(fout, "1 1 1 %4s %4s %4s %c ", sf.H[i], sf.K[i], sf.L[i], flag[i]);

            other_to_f(i, F, Fs) ;
            fprintf(fout, " %8.1f %5.1f \n" ,   F, Fs);
            
        }else{
            fprintf(fout, " ?  ?  ?  ?  ?  ?  ? \n" );
        }
        
        
    }
    
    fclose(fout);
}

           
void write_hkl_values(FILE *fout,FILE *finfo, char *fname, CifFile *fobjR, int i,int nd, char *pdb_id, int key)
// i= data block number ; nd=total number of blocks
{
    char outfile[256], block_name[256],  blockId[80],  tmp1[100],bid[80];
    vector<string> blockNames;
    int   index[20],nf=0, j;
    float cell[6], resol=50;
    FILE *fo=NULL;
    
    strcpy(bid, "_ABCDEFGHIJKLMNOPQRSTUVWXYZ123456789abcdefghijklmnopqrstuvwxyz");

    
    fobjR->GetBlockNames(blockNames);
    
    strcpy(blockId , blockNames[i].c_str());
    rid_of_space(blockId);
    if(strlen(blockId)==0){
        strcpy(blockId, "UNK");
    }
    
    parse_sf_values(fobjR, i);
    for (j=0; j<sf.nref; j++){//merge two free set to one!(mainly phenix I+,I-)
           
        if (sf.status && sf.free1){
            if (sf.status[j][0]=='?' && sf.free1[j][0]!='?'){
                strcpy(sf.status[j], sf.free1[j]);
            }
            
        }else if (sf.status && sf.free2 ){
            if(sf.status[j][0]=='?' && sf.free2[j][0]!='?'){
                strcpy(sf.status[j], sf.free2[j]);
            }
        }
           
    }
    if(sf.nref<=10) {
        printf("Error! Number of reflections is less than 10!\n");
        return;
    }
    
    if(sf.a && sf.b && sf.c && sf.alpha && sf.beta && sf.gamma ){  
        cell[0] = atof(sf.a);
        cell[1] = atof(sf.b);
        cell[2] = atof(sf.c);
        cell[3] = atof(sf.alpha);
        cell[4] = atof(sf.beta);
        cell[5] = atof(sf.gamma);
    }else if(CELL[0]>0.01 && CELL[1]>0.01 && CELL[2]>0.01 &&
             CELL[3]>0.01 && CELL[4]>0.01 && CELL[5]>0.01 ){  
        cell[0] = CELL[0];
        cell[1] = CELL[1];
        cell[2] = CELL[2];
        cell[3] = CELL[3];
        cell[4] = CELL[4];
        cell[5] = CELL[5];
    }
    if(sf.a && CELL[0]>0.01 &&
       (fabs(atof(sf.a) - CELL[0])>0.2 || fabs(atof(sf.b) - CELL[1])>0.2 ||
        fabs(atof(sf.alpha) - CELL[3])>0.5 || fabs(atof(sf.beta) - CELL[4])>0.5) ){
        strcpy(tmp1,"Error! Cells from SF and PDB don't match \n");
        write_sf_info(finfo, tmp1, 0);
    }

    get_num_file(sf.cryst_id, sf.wave_id, sf.scale_id, sf.nref, index, &nf) ;
    nf=1;
    
//    printf("BlockID=%s; PDBID=%s; Block Number=%d; File Number =%d : Mode=%d\n",
//           blockId, pdb_id, i,nf, SINGLE);

    if(key==0 && i==0){
        sprintf(block_name, "r%ssf",pdb_id);
        sf_related_head(fout, block_name, pdb_id, i);
        write_sf_file(fout, block_name, 1, i, index, key);
        return;
    }
    
        
    if(SINGLE==0){//block by block
        sprintf(outfile, "SF_%s.mmcif",blockId);
        if((fo = fopen(outfile, "w"))==NULL){        
            printf("Error! I have no permision to write the file %s\n",outfile);
            sprintf(tmp1,"%.3f %s,", resol,outfile);
            strcat(fname, tmp1);
            return ;
        }

        if(i==0){ //first block
            sprintf(block_name, "r%ssf",pdb_id);
        }else{ //other block
            sprintf(block_name, "r%s%csf",pdb_id,bid[i]);
        }
        
        sf_related_head(fo, block_name, pdb_id, i);
        
        sprintf(tmp1,"\nSF information for data block : %s\n", block_name);
        write_sf_info(finfo, tmp1, 1);
        
        if(sf.Fo && !sf.Fo_au){
            check_sf_infor(finfo,sf.H, sf.K, sf.L, sf.Fo, sf.sFo, sf.F2o,sf.sF2o ,
                           sf.Io,sf.sIo, sf.I_plus,sf.F_plus, sf.status,cell, 0, index,&resol);
        }else{
            check_sf_infor(finfo,sf.H, sf.K, sf.L, sf.Fo_au, sf.sFo_au, sf.F2o,sf.sF2o ,
                           sf.Io,sf.sIo, sf.I_plus,sf.F_plus, sf.status,cell, 0, index,&resol);
        }
        
        
        write_sf_file(fo, block_name, 1, i, index, key);

        if(nd==1)
            fprintf(fo,"#END OF REFLECTIONS\n");
        else{
            if(i<nd-1)
                fprintf(fo,"#END\n");
            else
                fprintf(fo,"#END OF REFLECTIONS\n");
        }
        
        fclose(fo);
//        printf("\nThe output file=%s :: resh=%.2f\n",outfile, resol);
        sprintf(tmp1,"%.3f %s,", resol,outfile);
        strcat(fname, tmp1);
        
//        if(key==0 && i==0) return;
        
    }else{ // parse all blocks
        for(j=0; j<nf; j++){
            
            if(j==0){
                if(strlen(pdb_id)>0){
                    sprintf(block_name, "%s_%d",pdb_id,j);
                    sf_related_head(fo, block_name, pdb_id, i);
                    
                }else{
                    sprintf(block_name, "%s_%d",blockId,j);
                    fprintf(fo, "data_%s\n",block_name);
                }
            }else{
                sprintf(block_name, "%s_%d",blockId,j);
                fprintf(fo, "data_%s\n",block_name);
            }
            
            
            printf("\nSF information for data block: %s : SET=%d\n", block_name, j+1);
            fprintf(finfo,"\nSF information for data block: %s : SET=%d\n", block_name, j+1);
            if(sf.cryst_id && sf.wave_id){
                printf("SF extracting (Crystal ID = %s ;  Wavelength ID = %s):\n",
                       sf.cryst_id[index[j]], sf.wave_id[index[j]]);
            
                sprintf(outfile, "SF_%s_%s_%s.cif",
                        sf.cryst_id[index[j]], sf.wave_id[index[j]], blockId);
            }else if(sf.cryst_id && !sf.wave_id){    
                printf("SF extracting (Crystal ID = %s):\n", sf.cryst_id[ index[j] ]);             
                sprintf(outfile, "SF_%s_%s_1.mmcif",blockId, sf.cryst_id[index[j]]);
                
            }else if(sf.wave_id && !sf.cryst_id){    
                printf("SF extracting (Wavelength ID = %s):\n", sf.wave_id[ index[j] ]);
                sprintf(outfile, "SF_%s_1_%s.mmcif",blockId, sf.wave_id[index[j]]);
            
            }else
                sprintf(outfile, "SF_%s_1_1.mmcif",blockId);
            
        
            if((fo = fopen(outfile, "w"))==NULL){        
                printf("I have no permision to write the file %s\n",outfile);
                return ;
            }
            
            if(sf.Fo && !sf.Fo_au){
                check_sf_infor(finfo,sf.H, sf.K, sf.L, sf.Fo, sf.sFo, sf.F2o,sf.sF2o ,
                               sf.Io,sf.sIo, sf.I_plus,sf.F_plus, sf.status,cell, j, index,&resol);
            }else{
                check_sf_infor(finfo,sf.H, sf.K, sf.L, sf.Fo_au, sf.sFo_au, sf.F2o,sf.sF2o ,
                               sf.Io,sf.sIo, sf.I_plus,sf.F_plus, sf.status,cell, j, index,&resol);
            }
            
            write_sf_file(fo, block_name, 2, j, index, key);

            fprintf(fo,"#END OF REFLECTIONS\n");
//    fprintf(fout,"#END\n");
            
            fclose(fo);
            printf("The output file=%s %.2f\n",outfile, resol);
        
            sprintf(tmp1,"%.3f %s,", resol,outfile);
            strcat(fname, tmp1);
        }
    }
    
    
}

void get_num_file(char **cryst_id, char **wave_id, char **scale_id,
                  int n, int *index, int *nf)
// n is # of reflections for the data-block.
// index  corresponding to HKL number (index[0]=1; index[1]=n)
{
    int i=0,  m1=0;

    if(!cryst_id){
        cryst_id = new char *[n+1];
        for(i=0; i<n; i++){
            cryst_id[i]=new char[2];
            strcpy(cryst_id[i], "1");
        }
    }
    
    if(!wave_id){
        wave_id = new char *[n+1];
        for(i=0; i<n; i++){
            wave_id[i]=new char[2];
            strcpy(wave_id[i], "1");
        }
    }
    
    if(!scale_id){
        scale_id = new char *[n+1];
        for(i=0; i<n; i++){
            scale_id[i]=new char[2];
            strcpy(scale_id[i], "1");
        }
    }
    num_file(cryst_id, wave_id, scale_id, n, index, &m1) ;
    *nf=m1;
}

void num_file(char **cryst_id, char **wave_id,char **scale_id, int n, int *index, int *nf) 
//get total different reflection files (nf,index)
{
    long i=0,  m1;
    char  **key=NULL, tmp[80], tmp1[10], tmp2[10], tmp3[10];

    
    key = new char *[n+1];
    
    
    (cryst_id) ? strcpy(tmp1,cryst_id[0]) :  strcpy(tmp1,"");
    (wave_id) ? strcpy(tmp1,wave_id[0]) :  strcpy(tmp2,"");
    (scale_id) ? strcpy(tmp1,scale_id[0]) :  strcpy(tmp3,"");
    sprintf(tmp,"%s_%s_%s", tmp1, tmp2, tmp3);
    
    key[0] = new char[strlen(tmp)+1];
    strcpy(key[0], tmp);

    
    m1=0;
    index[0]=0;
    for(i=1; i<n; i++){
        (cryst_id) ? strcpy(tmp1,cryst_id[i]) :  strcpy(tmp1,"");
        (wave_id) ? strcpy(tmp1,wave_id[i]) :  strcpy(tmp2,"");
        (scale_id) ? strcpy(tmp1,scale_id[i]) :  strcpy(tmp3,"");
        
        sprintf(tmp,"%s_%s_%s", tmp1, tmp2, tmp3);
        
        key[i] = new char[strlen(tmp)+1];
        strcpy(key[i], tmp);

        
        if(strcmp(tmp, key[i-1])){
            m1++;
            index[m1] = i;
        }
    }
    m1++;  // the last one
    index[m1] = n;
    *nf=m1;
    if(key)free_memory_2d(key, n);
}


void sf_related_head(FILE *fout_in, char *block_name, char *pdbid, int nd)
/* write all but not _refln */
{
    char tmp[256], tmp1[256], tmpname[256], **line, date[80], correct[512];
    int i, nline;
    FILE *fout;
    
    
    sprintf(tmpname, "data_%s_tmp.cif", block_name);
    fout=fopen(tmpname, "w");

    fprintf(fout, "data_%s\n",block_name);
    
    fprintf(fout,"#\n_entry.id  '%s'\n", pdbid);
    
    fprintf(fout,"#\n_exptl_crystal.id   1 \n");
    
    fprintf(fout,"#\n_reflns_scale.group_code   1 \n");

    strcpy(date,"");
    strcpy(correct,"");

        
    if (nd==0){
        write_audit(fout, date, correct, sf.naud, sf.aud_id, sf.aud_rec,
                    sf.aud_date, sf.aud_meth, 0);
    }
         
    if(sf.dif_cid ||  sf.dif_detail ||  sf.temp || sf.treat){
        write_diffrn(fout, sf.ndif, sf.dif_id, sf.dif_cid, sf.dif_detail, sf.temp, sf.treat, "");
    }
    

    if(sf.a &&  sf.b &&  sf.c && sf.alpha && sf.beta && sf.gamma){  
        fprintf(fout, "#\n_cell.entry_id  %s\n", pdbid);
        fprintf(fout, "_cell.length_a  %s\n", sf.a);
        fprintf(fout, "_cell.length_b  %s\n", sf.b);
        fprintf(fout, "_cell.length_c  %s\n", sf.c);
        fprintf(fout, "_cell.angle_alpha  %s\n", sf.alpha);
        fprintf(fout, "_cell.angle_beta  %s\n", sf.beta);
        fprintf(fout, "_cell.angle_gamma  %s\n", sf.gamma);
        
        delete []sf.a;delete []sf.b;delete []sf.c;
        delete []sf.alpha;delete []sf.beta;delete []sf.gamma;
    }

    if(sf.sg) {
        space_group_convertor(sf.sg, tmp);
        if(strlen(SYMM)>1){
            space_group_convertor(SYMM, tmp1);
            if (strcmp(tmp, tmp1)){
                printf("Error! The space group mismatch. sym_pdb='%s'; sym_sf='%s'\n", tmp1, tmp);
            }
        }
        
        fprintf(fout, "#\n_symmetry.entry_id  '%s'\n",pdbid);
        fprintf(fout, "_symmetry.space_group_name_H-M  '%s'\n",sf.sg);
        delete []sf.sg;
        
        if(sf.nsg)  {
            fprintf(fout, "_symmetry.Int_Tables_number  %s\n", sf.nsg);
            delete []sf.nsg;
        }
    }
    
    
    if(sf.nsymm){
        fprintf(fout, "#\nloop_\n");
        fprintf(fout, "_symmetry_equiv.id\n");
        fprintf(fout, "_symmetry_equiv.pos_as_xyz\n");
        for(i=0; i<sf.nsymm; i++){
            fprintf(fout,"%d  '%s'\n", i+1,  sf.symm[i]);
        }
    }
    
    if(sf.nwave){
        fprintf(fout, "#\nloop_\n");
        fprintf(fout, "_diffrn_radiation_wavelength.id \n");
        fprintf(fout, "_diffrn_radiation_wavelength.wavelength \n");
        for(i=0; i<sf.nwave; i++){
            fprintf(fout,"%d  %s \n", i+1,  sf.wavelength[i]);
        }
    }

   
    if(sf.resh or sf.nall or sf.nobs ){
        if (!sf.ndif){
            fprintf(fout, "#\n_diffrn.crystal_id   1 \n");
            fprintf(fout, "_diffrn.id   1 \n#\n");
        }
        
        fprintf(fout,"#\n_diffrn_reflns.diffrn_id  1\n");
        if(sf.resh) fprintf(fout,"_diffrn_reflns.pdbx_d_res_high  %s\n", sf.resh);
        if(sf.resl)fprintf(fout,"_diffrn_reflns.pdbx_d_res_low  %s\n", sf.resl);
        if(sf.hmax)fprintf(fout,"_diffrn_reflns.limit_h_max  %s\n", sf.hmax);
        if(sf.hmin)fprintf(fout,"_diffrn_reflns.limit_h_min  %s\n", sf.hmin);
        if(sf.kmax)fprintf(fout,"_diffrn_reflns.limit_k_max  %s\n", sf.kmax);
        if(sf.kmin)fprintf(fout,"_diffrn_reflns.limit_k_min  %s\n", sf.kmin);
        if(sf.lmax)fprintf(fout,"_diffrn_reflns.limit_l_max  %s\n", sf.lmax);
        if(sf.lmin)fprintf(fout,"_diffrn_reflns.limit_l_min  %s\n", sf.lmin);
        if(sf.nall)fprintf(fout,"_diffrn_reflns.number  %s\n", sf.nall);
        if(sf.nobs)fprintf(fout,"_diffrn_reflns.pdbx_number_obs  %s\n", sf.nobs);
    }

    
    fclose(fout);
    
    cifparse(tmpname, tmpname, 100) ;
    line=get_lines_from_file(tmpname, &nline);
    
    for (i=0; i<nline; i++){
        fprintf(fout_in, "%s", line[i]);
    }
    
    delete_file(tmpname);
    
    
}

void write_sf_file(FILE *fout,  char *block_name, int key, int i, int *index, int id)
//write all SF content
// key==1 write all in each data_block in one file;
// key==2 parse each separate set in each data_block & write them in each file   
{
    char flag=' ';
    long n, n1, nstart=0;
    float F=0, sF=0;
    
    if(key==2){ //each single set
        nstart=index[i];
        n1=index[i+1];
    }else{  // all sets in the block
        nstart=0;
        n1=sf.nref;
    }
    

///////////////
    

// never change the order of conditions.
    if(n1>10){
        fprintf(fout,"#\nloop_\n");
        fprintf(fout,"_refln.crystal_id\n");
        fprintf(fout,"_refln.wavelength_id\n");
        fprintf(fout,"_refln.scale_group_code\n");
            //fprintf(fout,"_refln.pdbx_diffrn_id\n");

        if(sf.H)fprintf(fout,"_refln.index_h\n");
        if(sf.K)fprintf(fout,"_refln.index_k\n");
        if(sf.L)fprintf(fout,"_refln.index_l\n");

//        if(sf.status || TEST>0)fprintf(fout,"_refln.status\n");
        fprintf(fout,"_refln.status\n");
        
        if(sf.Fo_au || sf.Fo)fprintf(fout,"_refln.F_meas_au\n"); 
        if(sf.sFo_au || sf.sFo)fprintf(fout,"_refln.F_meas_sigma_au\n");
        if(sf.Ic)fprintf(fout,"_refln.intensity_calc\n");
        if(sf.Io)fprintf(fout,"_refln.intensity_meas\n");
        if(sf.sIo)fprintf(fout,"_refln.intensity_sigma\n");
        
        if(sf.Fc)fprintf(fout,"_refln.F_calc\n");
        if(sf.Fc_au)fprintf(fout,"_refln.F_calc_au\n");
        if(sf.F2o)fprintf(fout,"_refln.F_squared_meas\n");
        if(sf.F2c)fprintf(fout,"_refln.F_squared_calc\n");
        if(sf.sF2o)fprintf(fout,"_refln.F_squared_sigma\n");
        
        if(sf.phase_c)fprintf(fout,"_refln.phase_calc\n");
        if(sf.phase_o)fprintf(fout,"_refln.phase_meas\n");
        
        if(sf.I_plus)fprintf(fout,"_refln.pdbx_I_plus\n");
        if(sf.sI_plus)fprintf(fout,"_refln.pdbx_I_plus_sigma\n");
        if(sf.I_minus)fprintf(fout,"_refln.pdbx_I_minus\n");
        if(sf.sI_minus)fprintf(fout,"_refln.pdbx_I_minus_sigma\n");
        
        if(sf.F_plus)fprintf(fout,"_refln.pdbx_F_plus\n");
        if(sf.sF_plus)fprintf(fout,"_refln.pdbx_F_plus_sigma\n");
        if(sf.F_minus)fprintf(fout,"_refln.pdbx_F_minus\n");
        if(sf.sF_minus)fprintf(fout,"_refln.pdbx_F_minus_sigma\n");
        
        if(sf.hla || sf.hla1)fprintf(fout,"_refln.pdbx_HL_A_iso\n");
        if(sf.hlb || sf.hlb1)fprintf(fout,"_refln.pdbx_HL_B_iso\n");
        if(sf.hlc || sf.hlc1)fprintf(fout,"_refln.pdbx_HL_C_iso\n");
        if(sf.hld || sf.hld1)fprintf(fout,"_refln.pdbx_HL_D_iso\n");

        if(sf.anom)fprintf(fout,"_refln.pdbx_anom_difference\n");
        if(sf.anoms)fprintf(fout,"_refln.pdbx_anom_difference_sigma\n");

        if(id==0 && !sf.Fo_au && !sf.Fo && (sf.I_plus || sf.F_plus || sf.Io)){
            fprintf(fout,"_refln.F_meas_au\n"); 
            fprintf(fout,"_refln.F_meas_sigma_au\n"); 
        }

        if(sf.d_spacing)fprintf(fout,"_refln.d_spacing\n");
        if(sf.a_calc)fprintf(fout,"_refln.A_calc\n");
        if(sf.b_calc)fprintf(fout,"_refln.B_calc\n");
        if(sf.weighted)fprintf(fout,"_refln.pdbx_FWT\n");
        if(sf.fs_uncorrected)fprintf(fout,"_refln.F_meas_sigma_uncorrected\n");
        if(sf.f_uncorrected)fprintf(fout,"_refln.F_meas_uncorrected\n");
        if(sf.gsas_i100_meas)fprintf(fout,"_refln.gsas_i100_meas\n");
        if(sf.i_unknown)fprintf(fout,"_refln.intensity_meas_unknown\n");
        if(sf.is_unknown)fprintf(fout,"_refln.intensity_sigma_unknown\n");
        if(sf.phase_part)fprintf(fout,"_refln.phase_part\n");
        if(sf.F_part_au)fprintf(fout,"_refln.F_part_au\n");
        
//        if(sf.Fc_all)fprintf(fout,"_refln.pdbx_F_calc_with_solvent\n");
//        if(sf.Fc_all_ph)fprintf(fout,"_refln.pdbx_phase_calc_with_solvent\n");
        if(sf.fwt)fprintf(fout,"_refln.pdbx_FWT\n");
        if(sf.phwt)fprintf(fout,"_refln.pdbx_PHWT\n");
        if(sf.delfwt)fprintf(fout,"_refln.pdbx_DELFWT\n");
        if(sf.delphwt)fprintf(fout,"_refln.pdbx_DELPHWT\n");
        
        if(sf.fom)fprintf(fout,"_refln.fom\n");
        
        if(sf.fib_co)fprintf(fout,"_refln.pdbx_fiber_coordinate\n");
        if(sf.fib_fo)fprintf(fout,"_refln.pdbx_fiber_F_meas_au\n");
        if(sf.fib_ly)fprintf(fout,"_refln.pdbx_fiber_layer\n");

        
    }
    
    
    for(n=nstart; n<n1; n++){
            //if(sf.status){ if(sf.status[n][0]=='x' || sf.status[n][0]=='X')continue;}
        
        (sf.cryst_id)? fprintf(fout, "%s ", sf.cryst_id[n]):fprintf(fout, "1 ");
        (sf.wave_id)? fprintf(fout, "%s ", sf.wave_id[n]):fprintf(fout, "1 ");
        (sf.scale_id)? fprintf(fout, "%s ", sf.scale_id[n]):fprintf(fout, "1 ");
            //(sf.diffr_id)? fprintf(fout, "%s ", sf.diffr_id[n]):fprintf(fout, "1 ");

        if(sf.H) fprintf(fout,"%4s ", sf.H[n]);
        if(sf.K) fprintf(fout,"%4s ", sf.K[n]);
        if(sf.L) fprintf(fout,"%4s ", sf.L[n]);

/*        
        if(sf.status || TEST>0) {
            if(TEST>0){
                (rand()%TEST == 1)? flag='f':flag='o';
                fprintf(fout,"%c ",flag);
                
            }else if(sf.status){
                fprintf(fout,"%1s ",sf.status[n]);
            }
            
        }  
*/        
        if(TEST>0){
            (rand()%TEST == 1)? flag='f':flag='o';
            fprintf(fout,"%c ",flag);
                
        }else if(sf.status){
            if(sf.Fo_au && sf.sFo_au && atof(sf.Fo_au[n])<0.0001 && strstr(sf.sFo_au[n], "999")){
                fprintf(fout," x ");
            }else{
                fprintf(fout,"%1s ",sf.status[n]);
            }
            
        }else{
            if(sf.Fo_au && sf.sFo_au && atof(sf.Fo_au[n])<0.0001 && strstr(sf.sFo_au[n],"999")){
                fprintf(fout," x ");
            }else{
                fprintf(fout," o ");
            }
        }
           
        if(sf.Fo_au) {
            if(strchr(sf.Fo_au[n], 'E'))
               fprintf(fout,"%8.2f ",atof( sf.Fo_au[n]));
            else
                fprintf(fout,"%8s ", sf.Fo_au[n]);
        }else if (sf.Fo) {
            if(strchr(sf.Fo[n], 'E'))
               fprintf(fout,"%8.2f ",atof( sf.Fo[n]));
            else
                fprintf(fout,"%8s ", sf.Fo[n]);
        }
        
        
        if(sf.sFo_au){
            if(strchr(sf.sFo_au[n], 'E'))
               fprintf(fout,"%6.2f ",atof( sf.sFo_au[n]));
            else
                fprintf(fout,"%6s ",  sf.sFo_au[n]);
            
        }else if(sf.sFo){
            if(strchr(sf.sFo[n], 'E'))
               fprintf(fout,"%6.2f ",atof( sf.sFo[n]));
            else
                fprintf(fout,"%6s ",  sf.sFo[n]);
        }
        
        
        if(sf.Ic){
            if(strchr(sf.Ic[n], 'E'))
               fprintf(fout,"%9.2f ",atof(sf.Ic[n]));
            else
                fprintf(fout,"%9s ",sf.Ic[n]);
        }
        if(sf.Io){
            if(strchr(sf.Io[n], 'E'))
               fprintf(fout,"%9.2f ",atof(sf.Io[n]));
            else
                fprintf(fout,"%9s ",sf.Io[n]);
        }
        if(sf.sIo){
            if(strchr(sf.sIo[n], 'E'))
               fprintf(fout,"%8.2f ",atof(sf.Io[n]));
            else
                fprintf(fout,"%8s ", sf.sIo[n]);
        }
        
        if(sf.Fc){
            (strchr(sf.Fc[n], 'E'))? fprintf(fout,"%10.2f ",atof(sf.Fc[n])): fprintf(fout,"%10s ",sf.Fc[n]);
        }
        
        if(sf.Fc_au){
            (strchr(sf.Fc_au[n], 'E'))? fprintf(fout,"%10.2f ",atof(sf.Fc_au[n])): fprintf(fout,"%10s ",sf.Fc_au[n]);
        }
        
        if(sf.F2o){
            (strchr(sf.F2o[n], 'E'))? fprintf(fout,"%10.2f ",atof(sf.F2o[n])): fprintf(fout,"%10s ",sf.F2o[n]);
        }
        if(sf.F2c){
            (strchr(sf.F2c[n], 'E'))? fprintf(fout,"%10.2f ",atof(sf.F2c[n])): fprintf(fout,"%10s ",sf.F2c[n]);
        }
        
        if(sf.sF2o){
            (strchr(sf.sF2o[n], 'E'))? fprintf(fout,"%8.2f ",atof(sf.sF2o[n])): fprintf(fout,"%8s ",sf.sF2o[n]);
        }
        
        
        if(sf.phase_c)fprintf(fout,"%7s ", sf.phase_c[n]);
        if(sf.phase_o)fprintf(fout,"%7s ", sf.phase_o[n]);
        
        if(sf.I_plus){
            (strchr(sf.I_plus[n], 'E'))? fprintf(fout,"%10.2f ",atof(sf.I_plus[n])): fprintf(fout,"%10s ", sf.I_plus[n]);
        }
        
        if(sf.sI_plus){
            (strchr(sf.sI_plus[n], 'E'))? fprintf(fout,"%8.2f ",atof(sf.sI_plus[n])): fprintf(fout,"%8s ", sf.sI_plus[n]);
        }
        
        if(sf.I_minus){
            (strchr(sf.I_minus[n], 'E'))? fprintf(fout,"%10.2f ",atof(sf.I_minus[n])): fprintf(fout,"%10s ", sf.I_minus[n]);
        }
        if(sf.sI_minus){
            (strchr(sf.sI_minus[n], 'E'))? fprintf(fout,"%8.2f ",atof(sf.sI_minus[n])): fprintf(fout,"%8s ", sf.sI_minus[n]);
        }
        
        if(sf.F_plus){
            (strchr(sf.F_plus[n], 'E'))? fprintf(fout,"%10.2f ",atof(sf.F_plus[n])): fprintf(fout,"%10s ", sf.F_plus[n]);
        } 
        if(sf.sF_plus){
            (strchr(sf.sF_plus[n], 'E'))? fprintf(fout,"%8.2f ",atof(sf.sF_plus[n])): fprintf(fout,"%8s ", sf.sF_plus[n]);
        } 
        
        if(sf.F_minus){
            (strchr(sf.F_minus[n], 'E'))? fprintf(fout,"%10.2f ",atof(sf.F_minus[n])): fprintf(fout,"%10s ", sf.F_minus[n]);
        }
        if(sf.sF_minus){
            (strchr(sf.sF_minus[n], 'E'))? fprintf(fout,"%10.2f ",atof(sf.sF_minus[n])): fprintf(fout,"%10s ", sf.sF_minus[n]);
        }
        
//        printf("H=%s K=%s L=%s ; %s  %s \n", sf.H[n], sf.K[n], sf.L[n],sf.sF_plus[n], sf.F_minus[n] );
        
        if(sf.hla)
            fprintf(fout,"%7s ", sf.hla[n]);
        else if(sf.hla1)
            fprintf(fout,"%7s ", sf.hla1[n]);
        
        if(sf.hlb)
            fprintf(fout,"%7s ", sf.hlb[n]);
        else if(sf.hlb1)
            fprintf(fout,"%7s ", sf.hlb1[n]);
        
        if(sf.hlc)
            fprintf(fout,"%7s ", sf.hlc[n]);
        else if(sf.hlc1)
            fprintf(fout,"%7s ", sf.hlc1[n]);
        
        if(sf.hld)
            fprintf(fout,"%7s ", sf.hld[n]);
        else if(sf.hld1)
            fprintf(fout,"%7s ", sf.hld1[n]);

        if(sf.anom)fprintf(fout,"%7s ", sf.anom[n]);
        if(sf.anoms)fprintf(fout,"%7s ", sf.anoms[n]);
        

        if(id==0 && !sf.Fo_au && !sf.Fo && (sf.I_plus || sf.F_plus || sf.Io)){
            other_to_f(n, F, sF) ;
            fprintf(fout," %7.2f %6.2f ",F, sF);
            
        }
        
        if(sf.d_spacing)fprintf(fout,"%3s ", sf.d_spacing[n]);
        if(sf.a_calc)fprintf(fout,"%7s ", sf.a_calc[n]);
        if(sf.b_calc)fprintf(fout,"%7s ", sf.b_calc[n]);
        if(sf.weighted)fprintf(fout,"%7s ", sf.weighted[n]);
        if(sf.fs_uncorrected)fprintf(fout,"%7s ", sf.fs_uncorrected[n]);
        if(sf.f_uncorrected)fprintf(fout,"%7s ", sf.f_uncorrected[n]);
        if(sf.gsas_i100_meas)fprintf(fout,"%7s ", sf.gsas_i100_meas[n]);
        if(sf.i_unknown)fprintf(fout,"%7s ", sf.i_unknown[n]);
        if(sf.is_unknown)fprintf(fout,"%7s ", sf.is_unknown[n]);
        if(sf.phase_part)fprintf(fout,"%7s ", sf.phase_part[n]);
        if(sf.F_part_au)fprintf(fout,"%7s ", sf.F_part_au[n]);
        
//        if(sf.Fc_all)fprintf(fout,"%7s ", Fc_all[n]);
//        if(sf.Fc_all_ph)fprintf(fout,"%7s ", Fc_all_ph[n]);
        if(sf.fwt)fprintf(fout,"%7s ", sf.fwt[n]);
        if(sf.phwt)fprintf(fout,"%7s ", sf.phwt[n]);
        if(sf.delfwt)fprintf(fout,"%7s ", sf.delfwt[n]);
        if(sf.delphwt)fprintf(fout,"%7s ", sf.delphwt[n]);
            
        if(sf.fom)fprintf(fout,"%5s ", sf.fom[n]);

        if(sf.fib_co)fprintf(fout,"%7s ", sf.fib_co[n]);
        if(sf.fib_fo)fprintf(fout,"%7s ", sf.fib_fo[n]);
        if(sf.fib_ly)fprintf(fout,"%7s ", sf.fib_ly[n]);

           
        fprintf(fout,"\n");
        
    }
    if(sf.dif_detail){
//        fprintf(fout,"#\n_diffrn.id  1 \n_diffrn.crystal_id 1\n" );
//        fprintf(fout,"#\n_diffrn.detail \n;%s;\n", sf.dif_detail[0]);
    }
    
   
}


void get_reso_hkl(char **status,char **H, char **K, char **L,float *cell, long n, float *resH)
{
    char shkl_max[80], shkl_min[80], hkl_max[80], hkl_min[80];
    
    int i, ah,ak,al,key=0;
    int max_H=-500, min_H=500, max_K=-500, min_K=500, max_L=-500, min_L=500;
    float max_R=0, min_R=500, rcell[6],resolution=0;
    
    if(cell[0]>0 && cell[1]>0){
        key=1;
        calc_recip_cell(cell, rcell);
    }

    for(i=0; i<n; i++){
        if(status){
            if(status[i][0]=='x' || status[i][0]=='X' ) continue;
        }
        
        ah = atoi(H[i]);
        ak = atoi(K[i]);
        al = atoi(L[i]);

        if(ah<min_H) min_H=ah;
        if(ak<min_K) min_K=ak;
        if(al<min_L) min_L=al;
       
        if(ah>max_H) max_H=ah;
        if(ak>max_K) max_K=ak;
        if(al>max_L) max_L=al;

//        if(key>0 && status[i][0]!='x'){
        if(key>0 ){
            
             resolution = get_resolution(ah, ak, al, rcell);
            if(resolution<min_R){
                min_R=resolution;
                sprintf(hkl_min, "%4d %4d %4d", ah, ak, al);
            }
            if(resolution>max_R){
                max_R=resolution;
                sprintf(hkl_max, "%4d %4d %4d", ah, ak, al);
            }
        } 
    }
    *resH=min_R;
    
    sprintf(shkl_max, "MIN_HKL=%s : RESOL=%.2f", hkl_max, max_R);
    sprintf(shkl_min, "MAX_HKL=%s : RESOL=%.2f", hkl_min, min_R);
    printf("%s\n", shkl_max );
    printf("%s\n", shkl_min );
    

}

void write_sf_info(FILE *fout, char *info, int id)
{
    if (id==0){
        fprintf(fout, "(%s) %s", PDBID, info);
        printf("(%s) %s", PDBID,info);
    }else if (id==1){
        fprintf(fout, "(%s) %s",PDBID, info);
    }else if (id==2){
        printf("(%s) %s", PDBID,info);
    }

}


int check_sf_infor(FILE *fout,char **H, char **K, char **L, char **Fo, char **sFo,
                   char **F2o,char **sF2o ,char **Io,char **sIo,char **I_plus,char **F_plus,
                   char **status, float *cell, int j, int *index, float *resol)
// check various errors in the single set
{
    char hkl_min[80], hkl_max[80], tmp[20], info[200];
    int ah,ak,al,key=0,err=-1, nstart=0, negsig=0;
    int max_H=-500, min_H=500, max_K=-500, min_K=500,max_L=-500, min_L=500;
    
    long  n_free=0, n_obs=0,nfpair=0, nref=0, nf_Fo=0, nf_F2o=0, nf_Io=0;
    long  n, i, n_1=0, n0=0, n1=0, n2=0, nf_sFo=0, nf_sIo=0;
    float f, sigf,  rcell[6], resolution;
    float max_F=-50000, min_F=50000, max_I=-50000, min_I=50000;
    float max_R=0, min_R=500, max_F2=-50000, min_F2=50000;
    float sum_i=0, sum_si=0, sum_f=0, sum_sf=0, i_over_si=0, f_over_sf=0;
    float sum_f2=0, sum_sf2=0, f2_over_sf2=0;


    *resol=50;
    
    
    nstart=index[j];
    n=index[j+1];

    if(!H || !K || !L ) {
        sprintf(info, "Error: The cif items 'index_h , index_k , index_l' are missing!\n");
        write_sf_info(fout, info, 0);
        err=0;
        return 0;
    }
    
    if(!Io && !Fo && !F2o &&!I_plus &&!F_plus ){
        sprintf(info, "Error: The cif items 'F_meas_au | intensity_meas | F_squared_meas' are missing!\n");
        write_sf_info(fout, info, 0);
        err=0;
        return 0;
    }
    
    if(n<50){
        sprintf(info, "Error: File conversion was wrong! Please check the data format.\n");
        write_sf_info(fout, info, 0);
    }
    
    if(cell[0]>0.01 && cell[1]>0.01){
        key=1;
        calc_recip_cell(cell, rcell);
    }
    
    for(i=nstart; i<n; i++){
        if (sFo && i>0 && atof(sFo[i-1])== atof(sFo[i])){
            nf_sFo++;
        }
        
        if (sIo && i>0 && atof(sIo[i-1])== atof(sIo[i])){
            nf_sIo++;
        }
        
        
        if((sFo && atof(sFo[i])<0.0) || (sIo && atof(sIo[i])<0.0)){
            negsig=1;
            continue;
        }
        if(Fo && sFo && atof(Fo[i])==-1.0 && atof(sFo[i])==0.0) continue;
        
        if(status){
            strcpy(tmp,status[i]);
            upperstr(tmp);
            
            if(tmp[0]=='X' || tmp[0]=='<' ||  tmp[0]=='H' ||  tmp[0]=='L' ||
               (tmp[0]=='-' and strlen(tmp)==1)) continue;
            nref++;
            
            if(tmp[0]=='O' )
                n_obs++;
            else if (tmp[0]=='F')
                n_free++;
            else if (atoi(tmp)<0)
                n_1++;
            else if (atoi(tmp)==0)
                n0++;
            else if (atoi(tmp) ==1)
                n1++;
            else if (atoi(tmp) ==2)
                n2++;
        }
        
        ah = atoi(H[i]);
        ak = atoi(K[i]);
        al = atoi(L[i]);

        if(ah<min_H) min_H=ah;
        if(ak<min_K) min_K=ak;
        if(al<min_L) min_L=al;
       
        if(ah>max_H) max_H=ah;
        if(ak>max_K) max_K=ak;
        if(al>max_L) max_L=al;

        if (F_plus){
            if(atof(F_plus[i])>0) nfpair++;
            if(atof(F_plus[i])>0) nfpair++;
        }else if (I_plus){
            if(atof(I_plus[i])>0) nfpair++;
            if(atof(I_plus[i])>0) nfpair++;
        }
        
            
        
        if(Fo) {
            f=atof(Fo[i]);
            if(f<min_F) min_F=f;
            if(f>max_F) max_F=f;
            if(sFo){
                sigf=atof(sFo[i]);
                if(sigf>0 && sFo[i][0] != '?'){
                    f_over_sf = f_over_sf + f/sigf;
                    sum_f = sum_f + f;
                    sum_sf = sum_sf + sigf;
                    nf_Fo++;
                }
            }
            
        }

        if(F2o) {
            f=atof(F2o[i]);
            if(f<min_F2) min_F2=f;
            if(f>max_F2) max_F2=f;
            if(sF2o){
                sigf=atof(sF2o[i]);
                if(sigf>0 && sF2o[i][0] != '?'){
                    f2_over_sf2 = f2_over_sf2 + f/sigf;
                    sum_f2 = sum_f2 + f;
                    sum_sf2 = sum_sf2 + sigf;
                    nf_F2o++;
                }
            }
        }
        
        if(Io) {
            f=atof(Io[i]);
            if(f<min_I) min_I=f;
            if(f>max_I) max_I=f;
            
            if(sIo){
                sigf=atof(sIo[i]);
                if(sigf>0 && sIo[i][0] != '?'){
                    i_over_si = i_over_si + f/sigf;
                    sum_i = sum_i + f;
                    sum_si = sum_si + sigf;
                    nf_Io++;
                    
                }
            }
        }
        
        if(key>0 ){
            
            resolution = get_resolution(ah, ak, al, rcell);
            if(resolution<min_R){
                min_R=resolution;
                sprintf(hkl_min, "%4d %4d %4d", ah, ak, al);
            }
            if(resolution>max_R){
                max_R=resolution;
                sprintf(hkl_max, "%4d %4d %4d", ah, ak, al);
            }
        }
    }

    if(negsig){
        sprintf(info,"Error:  The sigma value is negative. \n");
        write_sf_info(fout, info,0);
        err=1;
    }
    
    if(Fo && min_F<0){
        sprintf(info,"Error! The amplitude (F) is negative (I and F may be swapped).\n");
        write_sf_info(fout, info,0);
        err=1;
    }
    
    if(max_H>800 || max_K>800 || max_L>800 || min_H<-800 || min_K<-800 || min_L<-800){
        sprintf(info,"Error! The indices >800 or <-800. (columns might be connected)\n");
        write_sf_info(fout, info,0);
        err=1;
    }

    if(n>50 && nf_sFo>0 && n-nf_sFo<3){
        sprintf(info,"Warning! All the Sigma_Fo are the same!\n");
        write_sf_info(fout, info,0);
    }
    
    if(n>50 && nf_sIo>0 && n-nf_sIo<3){
        sprintf(info,"Warning! All the Sigma_Io are the same!\n");
        write_sf_info(fout, info,0);
    }
    
    
    if(status){
        
        if( strlen(FREERV)>0){ // manual input free R set
            sprintf(info,"Note: %s is used for free data set\n", FREERV);
            write_sf_info(fout, info,0);
            correct_status(n,status, 0);
            
            
        }else if( n_obs>0 && n_free>0 && n_obs < n_free ){
            sprintf(info,"Warning: Number of free set('f') > observed('o').(auto corrected)\n");
            write_sf_info(fout, info,0);
            correct_status(n,status, 1);
            err=1;
            
        }else if( nref==0 ){
            sprintf(info,"Warning: Status all assigned as 'x' or others. (auto assigned as 'o')\n");
            write_sf_info(fout, info,0);
            correct_status(n,status, 2);
            err=2;
            
        }else if(n_obs==0 && n_free==0 &&  n_1>10 && nref-n_1 > n_1 ){
            sprintf(info,"Warning: Status was assigned as '-1' for free set! (auto corrected)\n");
            write_sf_info(fout, info,0);
            correct_status(n,status, 3);
            err=3;
            
        }else if(n_obs==0 && n_free==0 &&  n_1==0  && n2==0 && n0>10 &&  nref-n0 > n0 ){
            sprintf(info,"Warning: Status was assigned as '0' for free set! (auto corrected)\n");
            write_sf_info(fout, info,0);
            correct_status(n,status, 4);
            err=4;
            
        }else if(n_obs==0 && n_free==0 &&  n_1==0 && n2==0 && n0>10 && n0>n1 ){
            sprintf(info,"Warning: Status was assigned as '1' for free set! (auto corrected)\n");
            write_sf_info(fout, info,0);
            correct_status(n,status, 5);
            err=5;
            
        }else if(n_obs==0 && n_free==0 &&  n_1==0 &&  n0>10 && n1>10 && n2>10 ){
            sprintf(info,"Warning: Status was assigned as numbers! (0 is auto-assigned as 'f')\n");
            write_sf_info(fout, info,0);
            correct_status(n,status,  6);
            err=6;
            
        }
        
    }
    
    if(cell[0]>0.001){
        sprintf(info,"Cell = %.2f %.2f %.2f %.2f %.2f %.2f\n", cell[0],cell[1],cell[2], cell[3],cell[4],cell[5]);
        write_sf_info(fout, info, 1);
    }
    
    if(key>0 && cell[0]>0.001){
        sprintf(info, "Lowest resolution= %7.2f ; corresponding HKL=%s\n",max_R,hkl_max);
        write_sf_info(fout, info, 1);
        sprintf(info, "Highest resolution=%7.2f ; corresponding HKL=%s\n",min_R,hkl_min);
        write_sf_info(fout, info, 1);
        *resol=min_R;
    }
    
    resolution = get_resolution(max_H, max_K, max_L, rcell);
    

    sprintf(info, "Max indices (Hmax=%4d  Kmax=%4d  Lmax=%4d) \n",max_H, max_K, max_L);
    write_sf_info(fout, info, 1);
    sprintf(info, "Min indices (Hmin=%4d  Kmin=%4d  Lmin=%4d) \n",min_H, min_K, min_L);
    write_sf_info(fout, info, 1);
    sprintf(info,"Total number of reflections in the SF file=%ld\n", n);
    write_sf_info(fout, info, 1);

    if(nfpair>100){
        sprintf(info, "Friedel pairs exist in SF file: Total number of F+ and F- =%ld\n", nfpair);
        write_sf_info(fout, info, 1);
    }
    
    if(sFo){
        sprintf(info,"Total number of observed reflections=%ld\n", nf_Fo);
        write_sf_info(fout, info, 1);
        
    }else if(sF2o){
        sprintf(info,"Total number of observed reflections=%ld\n", nf_F2o);
        write_sf_info(fout, info, 1);
    }else if(sIo){
        sprintf(info,"Total number of observed reflections=%ld\n", nf_Io);
        write_sf_info(fout, info, 1);
    }
    
    if(status){
        sprintf(info, "Number of observed  (status 'o')= %ld\n", n_obs);
        write_sf_info(fout, info, 1);
        sprintf(info, "Number for free set (status 'f')= %ld\n",n_free);
        write_sf_info(fout, info, 1);
        
        
        if(n_obs>0 && n_free>0 && n_free<n_obs){
            f=100*float(n_free)/(n_obs+n_free);
            sprintf(info,"Percentage for free set= %.2f\n",f);
            write_sf_info(fout, info, 1);
            
        }else if(n_obs>0 && n_free==0){
            sprintf(info,"Reflections have no free set!\n");
            write_sf_info(fout, info, 1);
        }
        
    }else{
        sprintf(info,"Reflections have no free set! (No status information)\n");
        write_sf_info(fout, info, 1);
    }

    if(Fo){
        sprintf(info,"maximum value of amplitude= %.2f\n",max_F);
        write_sf_info(fout, info, 1);
        
        sprintf(info,"minimum value of amplitude= %.2f\n",min_F);
        write_sf_info(fout, info, 1);

        sprintf(info,"<F/sigmaF> = %.2f;  <F>/<sigmaF> = %.2f; \n",  f_over_sf/nf_Fo, sum_f/sum_sf );
        write_sf_info(fout, info, 1);
        
    }
    
    if(F2o){
        sprintf(info,"maximum value of F square= %.2f\n",max_F2);
        write_sf_info(fout, info, 1);
        sprintf(info,"minimum value of F square= %.2f\n",min_F2);
        write_sf_info(fout, info, 1);
        sprintf(info,"<F2/sigmaF2> = %.2f;  <F2>/<sigmaF2> = %.2f; \n", f_over_sf/nf_F2o, sum_f/sum_sf );
        write_sf_info(fout, info, 1);
        
    }
    
    if(Io){
        sprintf(info,"maximum value of intensity= %.2f\n",max_I);
        write_sf_info(fout, info, 1);
        sprintf(info,"minimum value of intensity= %.2f\n",min_I);
        write_sf_info(fout, info, 1);
        sprintf(info,"<I/sigmaI> = %.2f;  <I>/<sigmaI> = %.2f; \n",i_over_si/nf_Io, sum_i/sum_si );
    }
    return err;
}

    
void correct_status(long n, char **status, int id)
{
    long i, nf=0;
    
    for(i=0; i<n; i++){
        if (id==0){ // Given status 
            if(atoi(status[i])==atoi(FREERV)){
                strcpy(status[i], "f");
                nf++;
            }else
                strcpy(status[i], "o");
            
        }else if (id==1){ //free set more than working set, swap
            if(strstr(status[i], "f"))
               strcpy(status[i], "o");
            else if(strstr(status[i], "o"))
                strcpy(status[i], "f");
            
        }else if (id==2){ // only given x
            if(strstr(status[i], "x"))
                strcpy(status[i], "o");
            
        }else if (id==3){ // -1 for free and others for 'o'
            if(atoi(status[i])<0)
                strcpy(status[i], "f");
            else
                strcpy(status[i], "o");
                
        }else if (id==4){ // 0 for 'f' and 1 for 'o'
            if(atoi(status[i])==0)
                strcpy(status[i], "f");
            else
                strcpy(status[i], "o");
            
        }else if (id==5){ // 0 for 'o' and 1 for 'f'
            if(atoi(status[i])==0)
                strcpy(status[i], "o");
            else
                strcpy(status[i], "f");
            
        }else if (id==6){ // 0 for 'f' and others for 'o'
            if(atoi(status[i])==0)
                strcpy(status[i], "f");
            else
                strcpy(status[i], "o");
            
        }
    }
    if (id==0 && nf==0){
        printf("Warning: free set (given as %s) not found. SF status is all given as 'o'\n",FREERV );
    }
    
        
}

        

float get_resolution(float h, float k, float l, float *rcell)
{

    float aa1, aa2, aa3, a2, b2, c2, resol=0, dist_sq;
    const float DEG2RAD= 0.017453292;
    
    
    aa1 =  2 * rcell[0] * rcell[1] * cos(DEG2RAD*rcell[5]);
    aa2 =  2 * rcell[0] * rcell[2] * cos(DEG2RAD*rcell[4]);
    aa3 =  2 * rcell[1] * rcell[2] * cos(DEG2RAD*rcell[3]);

    a2 =  rcell[0]*rcell[0];
    b2 =  rcell[1]*rcell[1];
    c2 =  rcell[2]*rcell[2];

    dist_sq = h*h*a2 + k*k*b2 + l*l*c2 + h*k*aa1 + h*l*aa2 + k*l*aa3;
    if(dist_sq>0.0000001)
        resol = sqrt(1.0/dist_sq);

    return resol;
    
}
    
void calc_recip_cell(float *cell, float *rcell)
{
    float cosa,cosb, cosc,sina,sinb,sinc, cosast,cosbst,coscst,v;
    const float DEG2RAD= 0.017453292;
    
    
    cosa = cos(DEG2RAD*cell[3]);
    cosb = cos(DEG2RAD*cell[4]);
    cosc = cos(DEG2RAD*cell[5]);
    sina = sin(DEG2RAD*cell[3]);
    sinb = sin(DEG2RAD*cell[4]);
    sinc = sin(DEG2RAD*cell[5]);
    
    v = cell[0]* cell[1] *cell[2]*
        sqrt(1. - cosa*cosa - cosb*cosb - cosc*cosc + 2.*cosa*cosb*cosc );
    
    rcell[0] =  cell[1] * cell[2]* sina / v;
    rcell[1] =  cell[0] * cell[2]* sinb / v;
    rcell[2] =  cell[0] * cell[1]* sinc / v;
    
    cosast   =  (cosb*cosc-cosa) / (sinb*sinc);
    cosbst   =  (cosa*cosc-cosb) / (sina*sinc);
    coscst   =  (cosa*cosb-cosc) / (sina*sinb);
    
    rcell[3] =  acos(cosast)/DEG2RAD;
    rcell[4] =  acos(cosbst)/DEG2RAD;
    rcell[5] =  acos(coscst)/DEG2RAD;
    
}



void get_mmcif_item_index(int n, char **item, int index[],  int *nitem_tot)  
//nitem=19, it should change when items change.

{
    int i;

    *nitem_tot=19;
    
    for(i=0; i<n; i++){
        if(!strcmp_case(item[i], "_refln.index_h")){
            index[0]=i;
        }else if(!strcmp_case(item[i], "_refln.index_k")){
            index[1]=i;
        }else if(!strcmp_case(item[i], "_refln.index_l")){
            index[2]=i;
        }else if(!strcmp_case(item[i], "_refln.F_meas_au")){
            index[3]=i;
        }else if(!strcmp_case(item[i], "_refln.F_meas_sigma_au")){
            index[4]=i;
        }else if(!strcmp_case(item[i], "_refln.intensity_meas")){
            index[5]=i;
        }else if(!strcmp_case(item[i], "_refln.intensity_sigma")){
            index[6]=i;
        }else if(!strcmp_case(item[i], "_refln.F_squared_meas")){
            index[7]=i;
        }else if(!strcmp_case(item[i], "_refln.F_squared_sigma")){
            index[8]=i;
        }else if(!strcmp_case(item[i], "_refln.status")){
            index[9]=i;
        }else if(!strcmp_case(item[i], "_refln.F_meas")){
            index[10]=i;
        }else if(!strcmp_case(item[i], "_refln.F_meas_sigma")){
            index[11]=i;
        }else if(!strcmp_case(item[i], "_refln.F_calc")){
            index[12]=i;
        }else if(!strcmp_case(item[i], "_refln.F_calc_au")){
            index[13]=i;
        }else if(!strcmp_case(item[i], "_refln.F_squared_calc")){
            index[14]=i;
        }else if(!strcmp_case(item[i], "_refln.intensity_calc")){
            index[15]=i;
        }else if(!strcmp_case(item[i], "_refln.fom")){
            index[16]=i;
        }else if(!strcmp_case(item[i], "_refln.wavelength_id")){
            index[17]=i;
        }else if(!strcmp_case(item[i], "_refln.crystal_id")){
            index[18]=i;
        }
        
    }
    
    *nitem_tot=19;
    
        /*
    for(i=0; i<n; i++){
        if(index[i]>=0)printf("cif-items %d  %d (%d) %s\n",
        n, i, index[i],item[index[i]]);
        printf("cif-items %d  %d (%d)\n",n, i, index[i]);
    }
        */    
}
void cif_token_change(const char *old_token, const char *new_token)
{
    printf("Warning! The mmcif token  _refln.%s is wrong!\n", old_token);
    printf("It has been corrected as _refln.%s \n",new_token);
}

void parse_sf_values(CifFile *fobjR, int i)
//parse all the sf values
{
    long n1=0,n2=0, j, k1=0, k2=0, naud=0, ndif=0 ;
    string diags;
    const char *Bid;
    vector<string> blockNames;
    
    
    fobjR->GetBlockNames(blockNames);
    Bid = blockNames[i].c_str();
    
    sf.a = get_attribute_value(fobjR, Bid, "cell", "length_a");
    sf.b = get_attribute_value(fobjR, Bid, "cell", "length_b");
    sf.c = get_attribute_value(fobjR, Bid, "cell", "length_c");
    sf.alpha = get_attribute_value(fobjR, Bid, "cell", "angle_alpha");
    sf.beta = get_attribute_value(fobjR, Bid, "cell", "angle_beta");
    sf.gamma = get_attribute_value(fobjR, Bid, "cell", "angle_gamma");
    sf.sg = get_attribute_value(fobjR, Bid, "symmetry", "space_group_name_H-M");
    sf.nsg = get_attribute_value(fobjR, Bid, "symmetry", "Int_Tables_number");

    sf.resh=get_attribute_value(fobjR, Bid,"reflns", "d_resolution_high");
    if (!sf.resh) sf.resh=get_attribute_value(fobjR, Bid,"diffrn_reflns", "pdbx_d_res_high");
        
    sf.resl=get_attribute_value(fobjR, Bid,"reflns", "d_resolution_low");
    if (!sf.resl) sf.resl=get_attribute_value(fobjR, Bid,"diffrn_reflns", "pdbx_d_res_low");
    
    sf.hmax=get_attribute_value(fobjR, Bid,"reflns", "limit_h_max");
    if (!sf.hmax) sf.hmax=get_attribute_value(fobjR, Bid,"diffrn_reflns", "limit_h_max");

    sf.kmax=get_attribute_value(fobjR, Bid,"reflns", "limit_k_max");
    if (!sf.kmax) sf.kmax=get_attribute_value(fobjR, Bid,"diffrn_reflns", "limit_k_max");
    
    sf.lmax=get_attribute_value(fobjR, Bid,"reflns", "limit_l_max");
    if (!sf.lmax) sf.lmax=get_attribute_value(fobjR, Bid,"diffrn_reflns", "limit_l_max");
    
    sf.hmin=get_attribute_value(fobjR, Bid,"reflns", "limit_h_min");
    if (!sf.hmin) sf.hmin=get_attribute_value(fobjR, Bid,"diffrn_reflns", "limit_h_min");

    sf.kmin=get_attribute_value(fobjR, Bid,"reflns", "limit_k_min");
    if (!sf.kmin) sf.kmin=get_attribute_value(fobjR, Bid,"diffrn_reflns", "limit_k_min");

    sf.lmin=get_attribute_value(fobjR, Bid,"reflns", "limit_l_min");
    if (!sf.lmin)sf.lmin =get_attribute_value(fobjR, Bid,"diffrn_reflns", "limit_l_min");
    
    sf.nall=get_attribute_value(fobjR, Bid,"reflns", "number_all");
    if (!sf.nall) sf.nall=get_attribute_value(fobjR, Bid,"diffrn_reflns", "number");

    sf.nobs=get_attribute_value(fobjR, Bid,"reflns", "number_obs");
    if (!sf.nobs) sf.nobs=get_attribute_value(fobjR, Bid,"diffrn_reflns", "pdbx_number_obs");

    
    
    sf.wavelength = get_attribute_values(fobjR, Bid, "diffrn_radiation_wavelength", "wavelength", &k1);
    sf.symm=get_attribute_values(fobjR, Bid,"symmetry_equiv", "pos_as_xyz", &k2);

    sf.aud_id=get_attribute_values(fobjR, Bid,"audit", "revision_id", &naud);
    sf.aud_date=get_attribute_values(fobjR, Bid,"audit", "creation_date", &naud);
    sf.aud_rec=get_attribute_values(fobjR, Bid,"audit", "update_record", &naud);
    sf.aud_meth=get_attribute_values(fobjR, Bid,"audit", "creation_method", &naud);

    sf.dif_id=get_attribute_values(fobjR, Bid,"diffrn", "id", &ndif);
    sf.dif_cid=get_attribute_values(fobjR, Bid,"diffrn", "crystal_id", &ndif);
    sf.temp=get_attribute_values(fobjR, Bid,"diffrn", "ambient_temp", &ndif);
    sf.treat=get_attribute_values(fobjR, Bid,"diffrn", "crystal_treatment", &ndif);
    sf.dif_detail=get_attribute_values(fobjR, Bid,"diffrn", "details", &ndif);

    
    sf.H = get_attribute_values(fobjR, Bid, "refln", "index_h", &n2);
    sf.K = get_attribute_values(fobjR, Bid, "refln", "index_k", &n2);
    sf.L = get_attribute_values(fobjR, Bid, "refln", "index_l", &n2);
    
    sf.Fo = get_attribute_values(fobjR, Bid, "refln", "F_meas", &n1);
    sf.sFo = get_attribute_values(fobjR, Bid, "refln", "F_meas_sigma", &n1);

    sf.Fo_au = get_attribute_values(fobjR, Bid, "refln", "F_meas_au", &n1);
    sf.sFo_au = get_attribute_values(fobjR, Bid, "refln", "F_meas_sigma_au", &n1);
    
    sf.Fc = get_attribute_values(fobjR, Bid, "refln", "F_calc", &n1);
    sf.Fc_au = get_attribute_values(fobjR, Bid, "refln", "F_calc_au", &n1);
    

    sf.Io = get_attribute_values(fobjR, Bid, "refln", "intensity_meas", &n1);
    if(!sf.Io){
        sf.Io = get_attribute_values(fobjR, Bid, "refln", "intensity_meas_au", &n1);
        if(sf.Io) cif_token_change("intensity_meas_au", "intensity_meas" );
        
        if(!sf.Io){
            sf.Io = get_attribute_values(fobjR, Bid, "refln", "intensity", &n1);
            if(sf.Io) cif_token_change("intensity", "intensity_meas" ); 
        }
    }
    
    
    sf.sIo = get_attribute_values(fobjR, Bid, "refln", "intensity_sigma", &n1);
    if(!sf.sIo){
        sf.sIo = get_attribute_values(fobjR, Bid, "refln", "intensity_sigma_au", &n1);
        if(sf.Io) cif_token_change("intensity_sigma_au", "intensity_sigma" );

        if(!sf.sIo){
            sf.sIo = get_attribute_values(fobjR, Bid, "refln", "intensity_sigm", &n1);
            if(sf.Io) cif_token_change("intensity_sigm", "intensity_sigma" );

            if(!sf.sIo){
                sf.sIo = get_attribute_values(fobjR, Bid, "refln", "intensity_meas_sigma", &n1);
                if(sf.Io) cif_token_change("intensity_meas_sigma", "intensity_sigma" );

                if(!sf.sIo){
                    sf.sIo = get_attribute_values(fobjR, Bid, "refln", "intensity_meas_sigma_au", &n1);
                    if(sf.Io) cif_token_change("intensity_meas_sigma_au", "intensity_sigma" );
                }
            }
            
        }
        
    }

    
    
    sf.F2o = get_attribute_values(fobjR, Bid, "refln", "F_squared_meas", &n1);
    sf.sF2o = get_attribute_values(fobjR, Bid, "refln", "F_squared_sigma", &n1);
    sf.Ic = get_attribute_values(fobjR, Bid, "refln", "intensity_calc", &n1);
    sf.F2c = get_attribute_values(fobjR, Bid, "refln", "F_squared_calc", &n1);
    
    sf.fom = get_attribute_values(fobjR, Bid, "refln", "fom", &n1);
    sf.phase_c = get_attribute_values(fobjR, Bid, "refln", "phase_calc", &n1);
    sf.phase_o = get_attribute_values(fobjR, Bid, "refln", "phase_meas", &n1);

    sf.status = get_attribute_values(fobjR, Bid, "refln", "status", &n1);
    if(!sf.status){
        sf.status = get_attribute_values(fobjR, Bid, "refln", "R_free_flag", &n1);
        if(sf.status) cif_token_change("R_free_flag", "status" );

        if(!sf.status){
            sf.status = get_attribute_values(fobjR, Bid, "refln", "statu", &n1);
            if(sf.status) cif_token_change("statu", "status" );

            if(!sf.status){
                sf.status = get_attribute_values(fobjR, Bid, "refln", "status_au", &n1);
                if(sf.status) cif_token_change("status_au", "status" );
            }
        }
    }
        //sf.free1 & sf.free2 special for two freeset, merge one to another if ?
    sf.free1 = get_attribute_values(fobjR, Bid, "refln", "R-free-flags(+)", &n1);
    sf.free2 = get_attribute_values(fobjR, Bid, "refln", "R-free-flags(-)", &n1);
    
        
    sf.wave_id = get_attribute_values(fobjR, Bid, "refln", "wavelength_id", &n1);
    sf.cryst_id = get_attribute_values(fobjR, Bid, "refln", "crystal_id", &n1);
    sf.scale_id = get_attribute_values(fobjR, Bid, "refln", "scale_group_code", &n1);
    
    sf.I_plus = get_attribute_values(fobjR, Bid, "refln", "pdbx_I_plus", &n1);
    sf.sI_plus = get_attribute_values(fobjR, Bid, "refln", "pdbx_I_plus_sigma", &n1);
    sf.I_minus = get_attribute_values(fobjR, Bid, "refln", "pdbx_I_minus", &n1);
    sf.sI_minus = get_attribute_values(fobjR, Bid, "refln", "pdbx_I_minus_sigma", &n1);
     
    sf.F_plus = get_attribute_values(fobjR, Bid, "refln", "pdbx_F_plus", &n1);
    sf.sF_plus = get_attribute_values(fobjR, Bid, "refln", "pdbx_F_plus_sigma", &n1);
    sf.F_minus = get_attribute_values(fobjR, Bid, "refln", "pdbx_F_minus", &n1);
    sf.sF_minus = get_attribute_values(fobjR, Bid, "refln", "pdbx_F_minus_sigma", &n1);
    
    sf.hla = get_attribute_values(fobjR, Bid, "refln", "pdbx_HL_A_iso", &n1);
    sf.hlb = get_attribute_values(fobjR, Bid, "refln", "pdbx_HL_B_iso", &n1);
    sf.hlc = get_attribute_values(fobjR, Bid, "refln", "pdbx_HL_C_iso", &n1);
    sf.hld = get_attribute_values(fobjR, Bid, "refln", "pdbx_HL_D_iso", &n1);

    sf.hla1 = get_attribute_values(fobjR, Bid, "refln", "pdbx_HLA", &n1);
    sf.hlb1 = get_attribute_values(fobjR, Bid, "refln", "pdbx_HLB", &n1);
    sf.hlc1 = get_attribute_values(fobjR, Bid, "refln", "pdbx_HLC", &n1);
    sf.hld1 = get_attribute_values(fobjR, Bid, "refln", "pdbx_HLD", &n1);

    sf.anom =  get_attribute_values(fobjR, Bid, "refln", "pdbx_anom_difference", &n1);
    if(!sf.anom){
        sf.anom =  get_attribute_values(fobjR, Bid, "refln", "DF_anomalous", &n1);
        if(sf.anom) cif_token_change("DF_anomalous", "pdbx_anom_difference" ); 
    }
    
    sf.anoms = get_attribute_values(fobjR, Bid, "refln", "pdbx_anom_difference_sigma", &n1);
    if(!sf.anoms){
        sf.anoms =  get_attribute_values(fobjR, Bid, "refln", "DF_anomalous_sigma", &n1);
        if(sf.anoms) cif_token_change("DF_anomalous_sigma", "pdbx_anom_difference_sigma" ); 
    }

    sf.d_spacing = get_attribute_values(fobjR, Bid, "refln", "d_spacing", &n1);
    sf.a_calc = get_attribute_values(fobjR, Bid, "refln", "A_calc", &n1);
    sf.b_calc = get_attribute_values(fobjR, Bid, "refln", "B_calc", &n1);
    sf.weighted = get_attribute_values(fobjR, Bid, "refln", "pdbx_fom_weighted_fmap", &n1);
    sf.fs_uncorrected = get_attribute_values(fobjR, Bid, "refln", "F_meas_sigma_uncorrected", &n1);
    sf.f_uncorrected = get_attribute_values(fobjR, Bid, "refln", "F_meas_uncorrected", &n1);
    sf.gsas_i100_meas = get_attribute_values(fobjR, Bid, "refln", "gsas_i100_meas", &n1);
    sf.i_unknown = get_attribute_values(fobjR, Bid, "refln", "intensity_meas_unknown", &n1);
    sf.is_unknown = get_attribute_values(fobjR, Bid, "refln", "intensity_sigma_unknown", &n1);
    sf.phase_part= get_attribute_values(fobjR, Bid, "refln", "phase_part", &n1);
    sf.F_part_au = get_attribute_values(fobjR, Bid, "refln", "F_part_au", &n1);

    sf.Fc_all = get_attribute_values(fobjR, Bid, "refln", "pdbx_F_calc_with_solvent", &n1);
    sf.Fc_all_ph = get_attribute_values(fobjR, Bid, "refln", "pdbx_phase_calc_with_solvent", &n1);
    sf.fwt = get_attribute_values(fobjR, Bid, "refln", "pdbx_FWT", &n1);
    sf.phwt = get_attribute_values(fobjR, Bid, "refln", "pdbx_PHWT", &n1);
    sf.delfwt = get_attribute_values(fobjR, Bid, "refln", "pdbx_DELFWT", &n1);
    sf.delphwt = get_attribute_values(fobjR, Bid, "refln", "pdbx_DELPHWT", &n1);

    sf.fib_co = get_attribute_values(fobjR, Bid, "refln", "pdbx_fiber_coordinate", &n2);
    sf.fib_fo = get_attribute_values(fobjR, Bid, "refln", "pdbx_fiber_F_meas_au", &n2);
    sf.fib_ly = get_attribute_values(fobjR, Bid, "refln", "pdbx_fiber_layer", &n2);

    
    sf.nref = n2;
    sf.nwave = k1;
    sf.nsymm = k2;
    sf.naud = naud;
    sf.ndif = ndif;
//    printf("Number of reflections for data set (%d) =%ld\n", i+1, n2);
    
    for(j=0; j<n2; j++){
//        printf("status=%s\n", sf.status[j]);
    }
    
                
}

